/* 
 * File:   Circuit.cpp
 * Author: Stanislav Pelák
 */

#include "Circuit.h"
#include "CNFmanager.h"
#include "SATsolverAdapter.h"
#include "Gene.h"
#include <iostream>
#include <list>
#include <sstream>
#include <stack>
#include <algorithm>
#include <assert.h>

#include <ctime>
using namespace std;

/**
 * creates Circuit object from Blif
 */
Circuit::Circuit(Blif* blif) {
    int geneIndex = 0;
    this->model = blif->GetModel();
    
    vector<string> inames = blif->getINames();
    vector<string> onames = blif->getONames();
    
    int inputs = inames.size();
    this->inputsCnt = inputs;
    
    //create inputs
    for(int i = 0; i < inputs; ++i) {
        int idx = geneIndex++;
        this->genes[idx] = new Gene(inames[i], idx);
    }
    
    //create nodes
    GateMapper* gMapper = GateMapper::getInstance();
    list<SoPLA>* data = blif->getData();
    for (list<SoPLA>::const_iterator iterator = data->begin(), end = data->end(); iterator != end; ++iterator) {
        int idx = geneIndex++;
        string id = (*iterator).getName();
        Gene* tmp = new Gene(id, idx);
        Gate* tmpg = gMapper->getGate((*iterator).getGateName());
        assert(tmpg != 0);
        tmp->SetGate(tmpg);
        this->genes[idx] = tmp;
        this->genesKeys.push_back(tmp->GetID());
    }
    
    //assign inputs to genes
    int geneIdx = this->inputsCnt;
    for (list<SoPLA>::const_iterator iterator = data->begin(), end = data->end(); iterator != end; ++iterator) {
        Gene* ref = this->genes[geneIdx];
        //assert(ref != 0);
        vector<string> inames = (*iterator).getINames();
        for(int i = 0; i < inames.size(); ++i) {
            string in = inames[i];
            Gene* inGene = getGeneById(in);
            ref->SetInput(i, inGene);
        }
        ++geneIdx;
    }
    //assert(geneIdx == this->genes.size());
    
    //assign outputs
    int outCnt = onames.size();
    vector<Gene*> out(outCnt);
    for(int i = 0; i < onames.size(); ++i) {
        string tmpOname = onames[i];
        Gene* outGene = getGeneById(tmpOname);
        out[i] = outGene;
    }
    this->outputs = out;
    this->fitness = -2;
}

Circuit::Circuit(const Circuit& orig) {
    this->copyGenotype(orig);
    this->model = orig.model;
    this->inputsCnt = orig.inputsCnt;
    this->fitness = orig.fitness;
}

Circuit::~Circuit() {
    int size = this->genes.size();
    for(int i = 0; i < size; ++i) {
        delete this->genes[i];
    }
}

void Circuit::copyGenotype(const Circuit& orig) {
    this->genesKeys = orig.genesKeys;
    this->inputsKeys = orig.inputsKeys;
    map<int, Gene*> origGenes = orig.genes;
    
    //copy genes
    int size = origGenes.size();
    for(int i = 0; i < size; ++i) {
        this->genes[i] = new Gene(*(origGenes[i % size]));
    }
    
    //assign outputs
    int outSize = orig.outputs.size();
    for(int i = 0; i < outSize; ++i) {
        string outID = orig.outputs[i]->GetID(); //todo index - faster
        Gene* outGene = this->getGeneById(outID);
        this->outputs.push_back(outGene);
    }
    
    //assign inputs to genes' gates
    this->assignInputs(&origGenes);
}

/**
 * Assigns inputs according to origGenes map
 * @param origGenes
 */
void Circuit::assignInputs(map<int, Gene*>* origGenes) {
    int genesSize = origGenes->size();
    Gene* orig;
    Gene* copied;
    Gene* tmpInput;
    for(int i = 0; i < genesSize; ++i) {
        orig = (*origGenes)[i];
        copied = this->genes[i];
        int inputsSize = (orig->GetInputs()).size();
        for(int j = 0; j < inputsSize; ++j) {
            int inputIdx = orig->GetInputs()[j]->GetIndex();
            copied->addInput(this->genes[inputIdx]);
        }
    }
}
    
vector<string> Circuit::getInputsKeys() {
    if(this->inputsKeys.size() > 0)
       return this->inputsKeys;
    vector<string> inIDs;
    for(int i = 0; i < this->inputsCnt; ++i)
        inIDs.push_back(this->genes[i]->GetID());
    this->inputsKeys = inIDs;
    return this->inputsKeys; 
}

vector<string> Circuit::getOutputsKeys() {
    if(this->outputsKeys.size() > 0)
       return this->outputsKeys; 
    int outCnt = this->outputs.size();
    for(int i = 0; i < outCnt; ++i) {
        this->outputsKeys.push_back(this->outputs[i]->GetID());
    }
    return this->outputsKeys; 
}

/**
 * returns string representation of Circuit's genotype
 */
string Circuit::getGenotype() {
    if(this->genotype.length() > 0)
        return this->genotype;
    ostringstream s;
    GateMapper* gmapper = GateMapper::getInstance();
    int inputs = this->inputsCnt;
    for(int i = inputs; i < this->genes.size(); ++i) {
        Gene* gene = this->genes[i];
        Gate* gate = gene->GetGate();
        string phenotypeFlag = "";
        if(gene->IsInPhenotype())
            phenotypeFlag = "+";
        s << "[" << gene->GetIndex() << "]" << phenotypeFlag << " " << gene->GetID() << "-" << gate->GetName() << "(" << gmapper->getCode(gate->GetName()) << ")|";
        for(int j = 0; j < gene->GetInputs().size(); ++j) {
            Gene* in = gene->GetInputs()[j];
            if(in) {
                s << in->GetID() << "|";
            } else {
                //s << "err|";
                assert(false);
            }
        }
    }
    s << "|";
    for(int o = 0; o < this->outputs.size(); ++o) {
        Gene* out = this->outputs[o];
        s << out->GetID() << "|";
    }
    string genotype(s.str());
    this->genotype = genotype;
    return genotype;
}

/**
 * conversion to CNF
 */
map<string, string> Circuit::GetCNF() {
    vector<Gene*> outputs = this->outputs;
    map<string, string> cnf;
    for(int i = 0; i < outputs.size(); ++i) {
        Gene* g = outputs[i];
        g->AppendGateCNF_rec(&cnf);
    }
    return cnf;
}

Gene* Circuit::getGeneById(string id) {
    Gene* tmp;
    for(int i = 0; i < this->genes.size(); ++i) {
        tmp = this->genes[i];
        if(tmp->GetID() == id)
            return tmp;
    }
    assert(false);
}

/**
 * returns vector of IDs of active nodes/genes
 */
vector<string> Circuit::getUtilizedGenesIDs() {
    int gsize = this->genes.size();
    for(int k = this->inputsCnt; k < gsize; ++k) {
        this->genes[k]->SetInPhenotype(false);
    }
    vector<string> keys;
    vector<Gene*> outputs = this->outputs;
    for(int i = 0; i < outputs.size(); ++i) {
        Gene* g = outputs[i];
        g->AppendGeneID_rec(&keys);
    }
    return keys;
}

bool Circuit::containsGene(string id) {
    Gene* tmp;
    for(int i = 0; i < this->genes.size(); ++i) {
        tmp = this->genes[i];
        if(tmp->GetID() == id)
            return true;
    }
    return false;
}

/**
 * checks equivalence to reference circuit. if fast=true, doesn't perform SAT based equivalence check if
 * the Circuit has already set fitness value
 */
bool Circuit::isEquivalent(Circuit* ref, bool fast) {
    if(fast && (this->fitness > -2)) {
        if(this->fitness == -1)
            return false;
        return true;
    }
    string dimacsCNF = CNFmanager::xorMergeToDIMACS(this, ref);
    bool sat = SATsolverAdapter::solve(dimacsCNF);
    if(sat) {
        this->fitness = -1;
        return false;
    }
    return true;
}

int Circuit::countFitness(int maxGates) {
    if(this->fitness > -2)
        return this->fitness;
    int utilized = this->getUtilizedGenesIDs().size();
    int fit = maxGates - utilized;
    this->fitness = fit;
    return fit;
}

Blif* Circuit::toBlif(vector<string> refOutputKeys) {
    //rename outputs if need to
    ostringstream s;
    vector<string> outputsKeys = this->getOutputsKeys();
    int out = outputsKeys.size();
    
    Blif* b = new Blif(this->getInputsKeys(), outputsKeys);
    
    vector<Gene*> outputs = this->outputs;
    map<string, SoPLA> pla;
    vector<string> keys;
    for(int i = 0; i < outputs.size(); ++i) {
        Gene* g = outputs[i];
        g->AppendGateSoPLA_rec(&pla, &keys);
    }
    
    int size = pla.size();
    for(int i = 0; i < size; ++i) {
        b->AppendPLA(&pla[keys[i]]);
    }
    return b;
}
