/* 
 * File:   Generation.cpp
 * Author: Stanislav Pelák
 */

#include "Generation.h"
#include "Randomizer.h"
#include "OutputManager.h"
#include "Configuration.h"
#include <ctime> 

string Generation::prevParent = "";
int Generation::maxFitness = -1;
MutableCircuit* Generation::bestCircuit = NULL;

Generation::Generation(Circuit* referenceCircuit, int r, int c, int l, int pos, int lambda, int mut_rate, int in, int out) {
    this->referenceCircuit = referenceCircuit;
    
    this->params["rows"] = r;
    this->params["columns"] = c;
    this->params["levels_back"] = l;
    this->params["population_size"] = lambda + 1;
    this->params["mutation_rate"] = mut_rate;
    this->params["inputs"] = in;
    this->params["outputs"] = out;
    
    this->generationID = 1;
    
    MutableCircuit* circuit;
    if(Configuration::GENERATE_GENOTYPE_RANDOMLY) {
        circuit = new MutableCircuit(referenceCircuit, r, c, l, pos, true);
    } else {
        circuit = new MutableCircuit(referenceCircuit, r, c, l, pos);
    }
    Generation::maxFitness = circuit->countFitness(r * c);
    Generation::prevParent = circuit->getGenotype();
    
    this->population.push_back(circuit);
    while(this->population.size() <= lambda) {
        MutableCircuit* next = new MutableCircuit(*circuit, false);
        next->mutate(mut_rate);
        if(!Configuration::EQUIV_CHILDREN || next->isEquivalent(this->referenceCircuit, true)) {
            this->population.push_back(next);
        } else {
            delete next;
        }
    }
}

/**
 * Edited copy of Generation constructor for testing totally random generation of individuals.
 */
Generation::Generation(Circuit* referenceCircuit, int r, int c, int l, int lambda, int mut_rate, int in, int out) {
    this->referenceCircuit = referenceCircuit;
    
    this->params["rows"] = r;
    this->params["columns"] = c;
    this->params["levels_back"] = l;
    this->params["population_size"] = lambda + 1;
    this->params["mutation_rate"] = mut_rate;
    this->params["inputs"] = in;
    this->params["outputs"] = out;
    
    this->generationID = 1;
    
    MutableCircuit* circuit;
    if(Configuration::GENERATE_GENOTYPE_RANDOMLY) {
        circuit = new MutableCircuit(referenceCircuit, r, c, l, 0, true);
    } else {
        circuit = new MutableCircuit(referenceCircuit, r, c, l, 0);
    }
    Generation::maxFitness = circuit->countFitness(r * c);
    Generation::prevParent = circuit->getGenotype();
    
    this->population.push_back(circuit);
    cout << endl << "Trying to generate equivalent circuit totally randomly" << endl;
    while(this->population.size() <= lambda) {
        MutableCircuit* next = new MutableCircuit(r, c, l, referenceCircuit->getInputsCnt(), referenceCircuit->getOutputs().size());
        if(!Configuration::EQUIV_CHILDREN || next->isEquivalent(this->referenceCircuit, true)) {
            this->population.push_back(next);
            cout << endl << "Equivalent circuit FOUND!" << endl;
        } else {
            delete next;
        }
    }
}

Generation::Generation(Circuit* referenceCircuit, MutableCircuit* individual, map<string, int> params, int id) {
    this->referenceCircuit = referenceCircuit;
    this->params = params;
    this->generationID = id;
    if((individual == 0) || (this->params.size() != 7)) {
        cerr << "Error(Generation:55): Invalid parameter(s)" << endl;
        exit(1);
    }
    MutableCircuit* copy = new MutableCircuit(*individual, true);
    copy->setFitness(individual->getFitness());
    this->population.push_back(copy);                          //insert parent
    Generation::prevParent = copy->getGenotype();
    
    while(this->population.size() < this->params["population_size"]) {
        MutableCircuit* circuit;
        if(!Configuration::RANDOM_INDIVIDUALS) {
            circuit = new MutableCircuit(*copy, false);
            circuit->mutate(this->params["mutation_rate"]);
        } else {
            circuit = new MutableCircuit(copy->getRows(), copy->getColumns(), copy->getColumns(), referenceCircuit->getInputsCnt(), referenceCircuit->getOutputs().size());
        }
        if(!Configuration::EQUIV_CHILDREN || circuit->isEquivalent(this->referenceCircuit, true)) {
            this->population.push_back(circuit);
        } else {
            delete circuit;
        }
    }
}

Generation::~Generation() {
    for(int i = 0; i < this->population.size(); ++i) {
        delete this->population[i];
    }
}

/**
 * Apply evolution on generation
 */
Generation* Generation::evolve(bool print) {
    vector<MutableCircuit*> possibleParents;
    MutableCircuit* prevParent;
    
    int cnt = this->population.size();  
    MutableCircuit* mc;
    
    for(int i = 0; i < cnt; ++i) {                      //test all individuals
        mc = this->population[i];
        bool equiv = true;
        if(!Configuration::EQUIV_CHILDREN) {
            equiv = mc->isEquivalent(this->referenceCircuit, true);  //if is equivalent => could be parent
        }
        if(equiv) {
            string genotype = mc->getGenotype();
            if(genotype == Generation::prevParent) {    // parent to this generation
                prevParent = mc;
            } else {
                possibleParents.push_back(mc);
            }
            
            int totalGenes = mc->getColumns() * mc->getRows();
            int fitness = mc->countFitness(totalGenes);
            if(fitness > Generation::maxFitness) {
                if(Generation::bestCircuit != 0)
                    delete Generation::bestCircuit;
                Generation::maxFitness = fitness;
                Generation::bestCircuit = new MutableCircuit(*mc, true);
                
                cout << "new max fitness: " << Generation::maxFitness << endl;
                cout << mc->getGenotype() << endl;
                
                if(print) {
                    ostringstream s;
                    s.str("");
                    s << totalGenes - fitness;
                    string gen(s.str());
                    OutputManager::printCircuitToBlif(mc, fitness, gen, this->referenceCircuit->getOutputsKeys());
                }
            }
        } else {
            assert(!Configuration::EQUIV_CHILDREN); //non-equivalent circuit in generation
        }
    }
    MutableCircuit* nextParent;
    if(possibleParents.size() > 0) {                                    //pick randomly from the best
        nextParent = Generation::pickRandom(&possibleParents);
    } else {
        nextParent = prevParent;
    }
    Generation* next = new Generation(this->referenceCircuit, nextParent, this->params, (this->generationID) + 1);
    return next;
}

MutableCircuit* Generation::pickRandom(vector<MutableCircuit*>* source) {
    srand((unsigned) time(0) + 151 * source->size() + Randomizer::getRandomizer()*9); 
    int random = rand();
    int idx = random % (source->size());
    return (*source)[idx];
}
