#include "stdafx.h"

#ifdef _MSC_VER
#pragma hdrstop
#endif

using namespace std;

#include "myExcept.h"
#include "bitvector.h"


bitvector::bitvector(void)
{
	bitsize = 0;
	data = NULL;
}

bitvector::bitvector( const unsigned int _size )
// Initializes the vector to 0
{
	if ( _size == 0 ) throw RuntimeException( "bitvector:: Zero vector size" );
	bitsize = _size;
	data = (DATATYPE *) calloc(items(), sizeof(DATATYPE));
	if ( data == NULL ) throw RuntimeException( "bitvector:: Not enough memory" );
}

bitvector::bitvector(const bitvector& _vector)
{
	if ( _vector.data != NULL ) {
		bitsize = _vector.bitsize;
		data = (DATATYPE *) malloc(_vector.items() * sizeof(DATATYPE));
		if ( data == NULL ) throw RuntimeException( "bitvector:: Not enough memory" );
		memcpy(data, _vector.data, _vector.items() * sizeof(DATATYPE));
	}
	else {
		bitsize = 0;
		data = NULL;
	}
}

bitvector::~bitvector(void)
{
	free(data);
}

bitvector& bitvector::operator=(const bitvector &_vector)
{
	if (this != &_vector) {
		free(data);
		bitsize = _vector.bitsize;
		data = (DATATYPE *) malloc(_vector.items() * sizeof(DATATYPE));
		if ( data == NULL ) throw RuntimeException( "bitvector:: Not enough memory" );
		memcpy(data, _vector.data, _vector.items() * sizeof(DATATYPE));
	}
	return *this;
}

void bitvector::resize( const unsigned int _size )
// resizes the vector to a given number of bits
{
	unsigned int olditems;

	if ( _size == 0 ) throw RuntimeException( "bitvector:: Zero vector size" );
	bitsize = _size;
	olditems = items();
	data = (DATATYPE *) realloc(data, items() * sizeof(DATATYPE));
	if ( data == NULL ) throw RuntimeException( "bitvector:: Not enough memory" );
	if (items() > olditems)
		memset(&data[olditems], 0, (items() - olditems) * sizeof(DATATYPE));
}

void bitvector::resize( const unsigned int _size, const bool val )
// resizes the vector and assignes a value to new vector items
{
	unsigned int olditems, oldsize;
	position p;

	if ( _size == 0 ) throw RuntimeException( "bitvector:: Zero vector size" );
	bitsize = _size;
	olditems = this->items();
	oldsize = this->size();
	data = (DATATYPE *) realloc(data, items() * sizeof(DATATYPE));
	if ( data == NULL ) throw RuntimeException( "bitvector:: Not enough memory" );

	if (items() > olditems) {
		GetPosition( oldsize, p );
		if ( p.bit != sizeof(DATATYPE)-1 ) {
			if ( val ) data[p.index] |= ( ALL_ONE << (p.bit+1) );
			else data[p.index] &= ~( ALL_ONE << (p.bit+1) );
		}
		memset(&data[olditems], val ? ALL_ONE : DATATYPE(0), (items() - olditems) * sizeof(DATATYPE));
	}
}

void bitvector::assign( const unsigned int _size, const bool val )
// assignes a value to all vector items
{
	DATATYPE d_val;

	if ( _size == 0 ) throw RuntimeException( "bitvector:: Zero vector size" );
	bitsize = _size;
	d_val = val ? ALL_ONE : DATATYPE(0);
	data = (DATATYPE *) realloc(data, items() * sizeof(DATATYPE));
	if ( data == NULL ) throw RuntimeException( "bitvector:: Not enough memory" );
	memset(data, d_val, items() * sizeof(DATATYPE));
}

inline void bitvector::GetPosition( const unsigned int & pos, position & result ) const
{
	result.index = (pos >> SHIFT);
	result.bit = pos & MASK;
}

void bitvector::push_back( const bool bit )
{
	position p;

	GetPosition( bitsize, p );	
	if ( p.bit == 0 ) {
		data = (DATATYPE *) realloc(data, (items() + 1) * sizeof(DATATYPE));
		if ( data == NULL ) throw RuntimeException( "bitvector:: Not enough memory" );
		data[items()] = DATATYPE(0);
	}
	bitsize++;
	if ( bit ) data[p.index] |= (1 << p.bit);
	else data[p.index] &= ~(1 << p.bit);
}

bool bitvector::operator[] ( const unsigned int pos ) const
{
	position p;

	if ( pos >= bitsize ) throw RuntimeException( "bitvector:: index out of range" );
	GetPosition(pos, p);
	return ((data[p.index] & (1 << p.bit)) != 0);
}

bool bitvector::operator== ( const bitvector &_vector ) const
{
	unsigned int i;
	position p;
	DATATYPE mask;

	if ( this->bitsize != _vector.bitsize ) return false;
	else {
		for ( i = 0; i < this->items()-1; i++ )
			if ( this->data[i] != _vector.data[i] ) return false;
		GetPosition(bitsize-1, p);
		mask = ~(ALL_ONE << (p.bit+1));
		if ( (this->data[i] & mask) != (_vector.data[i] & mask) ) return false;
	}
	return true;
}

void bitvector::clear(void)
{
	bitsize = 0;
}

void bitvector::setbit( const unsigned int pos, const bool val )
{
	position p;

	if ( pos > bitsize-1 ) throw RuntimeException( "bitvector:: index out of range" );
	GetPosition( pos, p );
	if ( !val ) data[p.index] &= ~(1 << p.bit);
	else data[p.index] |= (1 << p.bit);
}

bitvector bitvector::operator& ( const bitvector &_vector ) const
{
	bitvector vector;
	unsigned int i;

	if ( this->size() != _vector.size() ) throw RuntimeException( "bitvector:: & operator between vectors of different sizes" );
	vector.resize(this->size());
	for ( i = 0; i < this->items(); i++ ) vector.data[i] = this->data[i] & _vector.data[i];
	return vector;
}

bitvector bitvector::operator| ( const bitvector &_vector ) const
{
	bitvector vector;
	unsigned int i;

	if ( this->size() != _vector.size() ) throw RuntimeException( "bitvector:: & operator between vectors of different sizes" );
	vector.resize(this->size());
	for ( i = 0; i < this->items(); i++ ) vector.data[i] = this->data[i] | _vector.data[i];
	return vector;
}

bitvector bitvector::operator^ ( const bitvector &_vector ) const
{
	bitvector vector;
	unsigned int i;

	if ( this->size() != _vector.size() ) throw RuntimeException( "bitvector:: & operator between vectors of different sizes" );
	vector.resize(this->size());
	for ( i = 0; i < this->items(); i++ ) vector.data[i] = this->data[i] ^ _vector.data[i];
	return vector;
}

bitvector bitvector::operator~(void) const
{
	bitvector vector;
	unsigned int i;

	vector.resize(this->size());
	for ( i = 0; i < this->items(); i++ ) vector.data[i] = ~this->data[i];
	return vector;
}

bool bitvector::contains( const bitvector &_vector ) const
// returns true, if this contains _vector
// i.e., _vector does not have a true value, where this has
{
	unsigned int i;
	position p;
	DATATYPE mask;

	if ( this->size() != _vector.size() ) throw RuntimeException( "bitvector::contains: comparing vectors of different sizes" );
	for ( i = 0; i < this->items()-1; i++ )
		if ( (this->data[i] | _vector.data[i]) != this->data[i] ) return false;
	GetPosition(bitsize-1, p);
	mask = ~(ALL_ONE << (p.bit+1));
	if ( ((this->data[i] & mask) | _vector.data[i]) != (this->data[i] & mask) ) return false;
	return true;
}

bool bitvector::isNull(void) const
{
	unsigned int i;
	position p;
	DATATYPE mask;

	for ( i = 0; i < this->items()-1; i++ )
		if ( data[i] != 0 ) return false;
	GetPosition(bitsize-1, p);
	mask = ~(ALL_ONE << (p.bit+1));
	if ( (this->data[i] & mask) != 0 ) return false;
	return true;
}

bool bitvector::isOne(void) const
{
	unsigned int i;
	position p;
	DATATYPE mask;

	for ( i = 0; i < this->items()-1; i++ )
		if ( data[i] != ALL_ONE ) return false;
	GetPosition(bitsize-1, p);
	mask = ~(ALL_ONE << (p.bit+1));
	if ( (this->data[i] & mask) != mask ) return false;
	return true;
}

void bitvector::operator&= ( const bitvector &_vector )
{
	unsigned int i;

	if ( this->size() != _vector.size() ) throw RuntimeException( "bitvector:: & operator between vectors of different sizes" );
	for ( i = 0; i < this->items(); i++ ) this->data[i] &= _vector.data[i];
}

void bitvector::operator|= ( const bitvector &_vector )
{
	unsigned int i;

	if ( this->size() != _vector.size() ) throw RuntimeException( "bitvector:: & operator between vectors of different sizes" );
	for ( i = 0; i < this->items(); i++ ) this->data[i] |= _vector.data[i];
}

void bitvector::operator^= ( const bitvector &_vector )
{
	unsigned int i;

	if ( this->size() != _vector.size() ) throw RuntimeException( "bitvector:: & operator between vectors of different sizes" );
	for ( i = 0; i < this->items(); i++ ) this->data[i] ^= _vector.data[i];
}
