//---------------------------------------------------------------------------

#ifndef networkH
#define networkH

#include "Node.h"
#include "Signal.h"
   

class Blif;

struct Fault {		// for fault simulation
	string from;		// from gate output (name)
	string to;			// to gate input (name). Empty if no ambiguity
	bool stuck;			// stuck-at value
};

class Network {
  private:
	string model;			// name of the model
	list<Node> nodes;		// here are the PLAs
	vector<string> inames;			// names of input ports. Must be defined
	vector<string> onames;			// names of output ports. Must be defined

	list<Signal> signals;		// list of signals
	hashtable<list<Signal>::iterator> signalnames;	// hashtable of signal names. Must be kept compatible with "signals"!

	vector<list<Node>::iterator> indexnodes;	// vector of nodes, if Network is indexed
	
	list<Signal>::iterator FindOrCreateSignal( const string &name );

	vector<unsigned int> components;		// components generated by GetComponents().

	bool validlevels;		// are all the node levels valid?
	unsigned int levels;	// # of levels of the circuit. Valid only if validlevels = true

	void AddNodeSignals( list<Node>::iterator ndi );	// creates signal entries for a new node
	bool DeleteSignal( const string &n );
	void Disconnect(void);	// Deletes the connections

	double GetNodeDelay( list<Node>::iterator ndi, unsigned int model );
	void ComputeLevels(void);

  public:
	Network();
	Network( Network *net );
	Network( Blif &src );
	Network( PLA &src );
	Network(const vector<string> _inames, const vector<string> _onames );
	~Network();

	bool operator==( const Network& net );

	int tmp;				// for different purposes. Must not be modified by the kernel
	unsigned int int_tmp;	// for different purposes. May be modified by the kernel!

	// setters & getters
	list<Node> *getData(void) { return &nodes; };
	const list<Node> &getConstDataRef(void) const { return nodes; };
	unsigned int GetNodes(void) const { return nodes.size(); };		// returns number of nodes
	unsigned int inputs(void) const { return inames.size(); };
	unsigned int outputs(void) const { return onames.size(); };
	unsigned int getLiterals(void);
	void SetModel( const string _model ) { model = _model; };
	string getModel(void) const { return model; };
	list<Signal> *GetSignals(void);
	vector<string> getINames(void) const { return inames;};
	vector<string> getONames(void) const { return onames;};
	string getIName( const unsigned int i ) const;
	string getOName( const unsigned int i ) const;
	void setIName( const unsigned int i, const string name );
	void setOName( const unsigned int i, const string name );

	// info
	bool isPI( const string &name );		// is the signal a PI?
	bool isPO( const string &name );		// is the signal a PO?
	bool isConstant( const string &name );	// is the signal fed by a constant node?
	bool empty(void) const { return nodes.empty(); };
	pair<bool, list<Signal>::iterator> FindSignal( const string &n );
	bool IsSignalThere( const string &n );	// Does the signal with a given name exist in the network?
	bool IsGate( const string &n );			// Is the signal driven by some gate?
	unsigned int GetNodeLevel( list<Node>::iterator ndi );  // Returns the level of the node
	unsigned int GetLevels(void);			// returns the number of circuit levels
	unsigned int GetPIOrdinal( const string &name );		// Returns an ordinal value of a PI
	unsigned int GetPOOrdinal( const string &name );		// Returns an ordinal value of a PO
	unsigned int CountComponents(void);						// Returns the number of connected components

	// I/O
	void SaveMBlif( const char *fn );	// Saves the network to a MBLIF file

	// data manipulation
	list<Node>::iterator AddNode( const Node &_node );
	list<Node>::iterator AddNode( Node &_node );
	void SpliceNetworks( Network &net );			// splices two networks. Data from net is deleted
	list<Node>::iterator DeleteNode( const list<Node>::iterator ni );
	void ToSingleOutput(void);			// Converts the network into a network of single-output nodes, like Blif
	list<Node>::iterator MergeNodes( list<Node>::iterator node1, list<Node>::iterator node2 );  // merges two nodes into one
	void Rename( const string &oldname, const string &newname );	// Changes the given signal name, in the whole network
	void AddInput( const string &name );		// Adds an input
	void AddOutput( const string &name );		// Adds an output
	void DeleteInput( unsigned int i );			// Deletes i-th input
	void DeleteOutput( unsigned int i );			// Deletes i-th input

	void ReconnectInput( list<Node>::iterator ni, unsigned int pos, const string &name ); // Changes name of a given input
	void ReconnectOutput( list<Node>::iterator ni, unsigned int pos, const string &name ); // Changes name of a given output
	

	// conversion routines
	Blif *ToBlif(void) const;		// Converts the network to Blif
	Blif *ToBlif( const bool type ) const;		// Converts the network to Blif. type specifies the Blif type (false = r, true = f)

	// indexing
	void Reindex(void);				// creates the vector of nodes, for random access to nodes
	unsigned int getIndex(const list<Node>::iterator ndi);	// returns index of a given node. If not indexed, performs reindexation
	list<Node>::iterator GetNodeByIndex( const unsigned int i );		// returns an iterator to a node with a given index

	// other
	void DerivePorts(void);				// Generates PIs and POs by analyzing the network
	void DeleteVoidPorts(void);			// Deletes inputs and outputs not connected to any node
	list<string> FindVoidSignals(void);		// Returns a list of void (unfed) signals

	double GetDelay(unsigned int model);		// calculates the circuit delay
	double GetNCLDelay(void);		// calculates the circuit delay. To be deleted.
	double GetMinDelay(void);
	double GetAvgDelay(void);

	void GenerateSignals(void);		// Generates the signals structure

	vector<Literal> Simulate( const Cube &t, Fault *flt = NULL, bool all=false );		// simulate a vector

	vector<string> GetAllSignals(void);

	vector<unsigned int> *GetComponents(void)
		{ if (components.empty()) CountComponents(); return &components; };
	
	bool StructEqual( const Network& net );
};






//---------------------------------------------------------------------------
#endif
