#include "MutableCircuit.h"
#include <iostream>
#include <sstream>
#include <ctime> 
#include <cstdlib> 
#include <map>
#include <set>
#include <assert.h> 
#include "CNFmanager.h"
#include "SATsolverAdapter.h"
#include "Randomizer.h"
#include "Configuration.h"

using namespace std;

MutableCircuit::MutableCircuit(int nr, int nc, int l, int inputsCnt, int outputsCnt) {
    this->initEmptyCircuit(nr, nc, l, inputsCnt, outputsCnt);
    map<int, Gene*> g;
    this->fillGenes("in", inputsCnt, 0, g, &this->inputsKeys);
    this->fillGenes("g", nr*nc, inputsCnt, g, &this->genesKeys);
    this->genes = g;
    this->generateRandomGenotype();
    Configuration::nextCircuit();
}

MutableCircuit::MutableCircuit(const MutableCircuit& orig, bool copy) {
    this->rows = orig.rows;
    this->columns = orig.columns;
    this->l = orig.l;
    this->fitness = orig.getFitness();
    this->model = orig.model;
    this->inputsCnt = orig.inputsCnt;
    this->copyGenotype(orig);
    if(!copy) {
        Configuration::nextCircuit();
    }
}

MutableCircuit::MutableCircuit(Circuit* orig, int rows, int cols, int l, int pos) {
    assert((orig->getGenesKeys().size() + pos) <= cols);
    this->model = orig->getModel();
    this->fitness = orig->getFitness();
    this->initEmptyCircuit(rows, cols, l, orig->getInputsCnt(), orig->getOutputsKeys().size());
    this->generateCircuit(orig, rows, cols, pos);
}

MutableCircuit::~MutableCircuit() {
}

MutableCircuit::MutableCircuit(Circuit* orig, int rows, int cols, int l, int pos, bool randomEquiv) {
    vector<string> inputsKeys = orig->getInputsKeys();
    int inputsCnt = inputsKeys.size();
    this->inputsKeys = inputsKeys;
    
    map<int, Gene*> origGenes = orig->getGenes();
    int genesCnt = origGenes.size() - inputsCnt;
    assert((genesCnt + pos) <= cols);
            
    vector<Gene*> outputs = orig->getOutputs();
    this->initEmptyCircuit(rows, cols, l, inputsCnt, outputs.size());
    map<int, Gene*> g;
    for(int i = 0; i < inputsCnt; i++) {
        string id = inputsKeys[i];
        g[i] = new Gene(id, i);
    }
    this->fillGenes("g", rows*cols, inputsCnt, g, &this->genesKeys);
    this->genes = g;
    this->generateRandomGenotype();
    
    for(int idx = 0; idx < genesCnt; ++idx) {   //column index
        Gene* pattern = origGenes[idx + inputsCnt];
        Gene* toChange = this->genes[inputsCnt + (idx + pos)*rows];
        toChange->SetGate(pattern->GetGate());
        toChange->SetId(pattern->GetID());
        toChange->SetInPhenotype(true);
        vector<Gene*> inputs = pattern->GetInputs();        //vstupy genu
        int iSize = inputs.size();
        for(int i = 0; i < iSize; ++i) {
            Gene* inPat = inputs[i];
            int inPatIdx = inPat->GetIndex();
            int tmpIdx;
            if(inPatIdx < inputsCnt) {
                tmpIdx = inPatIdx;
            } else {
                int back = pattern->GetIndex() - inPat->GetIndex();
                assert(back > 0);
                tmpIdx = toChange->GetIndex() - (rows * back);
            }
            Gene* found = this->genes[tmpIdx];
            assert((found != 0) && (found->GetGate() == inPat->GetGate()) || (found->IsInPhenotype() && !inPat->IsInPhenotype()));
            toChange->SetInput(i, found);
        }
    }
    
    vector<string> outKeys = orig->getOutputsKeys();
    vector<Gene*> outGenes = orig->getOutputs();
    vector<Gene*> outg;
    int outSize = outGenes.size();
    for(int o = 0; o < outSize; ++o) {
        Gene* outPat = outGenes[o];
        int colIdx = outPat->GetIndex() - inputsCnt + pos;
        int tmpIdx = (colIdx * rows) + inputsCnt;
        Gene* out = this->genes[tmpIdx];
        assert(out->GetGate() == outPat->GetGate());
        out->SetId(outKeys[o]);
        outg.push_back(out);
    }
    this->outputs = outg;
    this->outputsKeys = outKeys;
    assert(this->isEquivalent(orig, false));
    //Configuration::nextCircuit();
}

void MutableCircuit::initEmptyCircuit(int nr, int nc, int l, int inputsCnt, int outputsCnt) {
    this->rows = nr;
    this->columns = nc;
    this->l = l;
    vector<Gene*> out (outputsCnt);
    this->outputs = out;
    this->inputsCnt = inputsCnt;
    this->fitness = -2;
}

/**
 * Generate MutableCircuit from reference Circuit with given phenotype size
 */
void MutableCircuit::generateCircuit(Circuit* circuit, int rows, int cols, int pos) {
    vector<string> inputsKeys = circuit->getInputsKeys();
    int in = inputsKeys.size();
    //fix input names
    for(int i = 0; i < in; ++i) {
        Gene* g = new Gene(inputsKeys[i], i);
        this->genes[i] = g;
    }
    this->inputsKeys = inputsKeys;
    this->inputsCnt = in;
    
    //fill genes
    int idx;
    vector<string> gKeys;
    map<int, Gene*> refGenes = circuit->getGenes();
    int circuitSize = refGenes.size() - in;
    string namePrefix = Configuration::NAME_PREFIX;
    ostringstream s;
    Gene* pattern;
    //cout << "start" << endl;
    for(int c = 0; c < cols; ++c) {
        int patternIdx = (c % circuitSize);
        if((c >= pos) && (c < (pos + circuitSize))) {
            patternIdx = (c - pos);
        }
        for(int r = 0; r < rows; ++r) {
            if(r == 0) {
                pattern = refGenes[patternIdx + in];
            } else {
                pattern = refGenes[(c % circuitSize) + in];
            }
            s.str("");
            idx = in + (c * rows) + r;          //index noveho genu
            s << namePrefix << idx;
            string id(s.str());                 //id noveho genu
            Gene* copy = new Gene(*pattern);
            copy->SetIndex(idx);
            copy->SetId(id);
            if((r == 0) && (c >= pos) && (c < (pos + circuitSize))) {
                copy->SetInPhenotype(true);
            } else {
                copy->SetInPhenotype(false);
            }
            vector<Gene*> inputs = pattern->GetInputs();        //vstupy genu
            int iSize = inputs.size();
            for(int i = 0; i < iSize; ++i) {
                Gene* inPat = inputs[i];
                int inPatIdx = inPat->GetIndex();
                int tmpIdx;
                if(inPatIdx < in) {
                    tmpIdx = inPatIdx;
                } else {
                    int back = pattern->GetIndex() - inPat->GetIndex();
                    assert(back > 0);
                    tmpIdx = copy->GetIndex() - (rows * back);
                }
                Gene* found = this->genes[tmpIdx];
                assert((found->GetGate() == inPat->GetGate()) || (found->IsInPhenotype() && !inPat->IsInPhenotype()));
                copy->addInput(found);
            }
            this->genes[idx] = copy;
            gKeys.push_back(id);
        }
    }
    this->genesKeys = gKeys;
    
    //set outputs
    vector<string> outKeys = circuit->getOutputsKeys();
    vector<Gene*> outGenes = circuit->getOutputs();
    vector<Gene*> outg;
    int outSize = outGenes.size();
    for(int o = 0; o < outSize; ++o) {
        Gene* outPat = outGenes[o];
        int colIdx = outPat->GetIndex() - in + pos;
        int tmpIdx = (colIdx * rows) + in;
        Gene* out = this->genes[tmpIdx];
        assert(out->GetGate() == outPat->GetGate());
        out->SetId(outKeys[o]);
        outg.push_back(out);
    }
    this->outputs = outg;
    this->outputsKeys = outKeys;
    assert(this->isEquivalent(circuit, false));
}

void MutableCircuit::fillGenes(string pref, int cnt, int offset, map<int, Gene*> &target, vector<string>* keys) {
    ostringstream s;
    for(int i = offset; i < cnt + offset; i++) {
        s.str("");
        s << pref << i;
        string id(s.str());
        target[i] = new Gene(id, i);
        if(keys != 0) {
            keys->push_back(id);
        }
    }
}

void MutableCircuit::generateRandomGenotype() {
    GateMapper* mapper = GateMapper::getInstance();
    /*Generate gates*/
    vector<string> genes = this->getGenesKeys();
    for(int i = this->inputsCnt; i < this->genes.size(); ++i) {
        Gate* gate = mapper->getRandomGate();
        Gene* gene = this->genes[i];
        gene->SetGate(gate);
        /*Generate inputs to each gene*/
        for(int in = 0; in < gate->GetInputs(); ++in) {
            Gene* input = generateInput(i);
            gene->SetInput(in, input);
        }
    }
    /*Generate outputs*/
    int genesCnt = this->genes.size();
    for(int i = 0; i < this->outputs.size(); ++i) {
        srand((unsigned) time(0) + i + Randomizer::getRandomizer()*22); 
        int random = rand();
        int idx = random % genesCnt;
        this->outputs[i] = this->genes[idx];
    }
}

Gene* MutableCircuit::generateInput(int geneIndex) {
    int inputs = this->inputsCnt;
    geneIndex -= inputs;
    int firstInColumn = geneIndex - (geneIndex % this->rows);
    int availableNodesBack = this->l * this->rows;
    if(availableNodesBack > (firstInColumn)) {
        availableNodesBack = firstInColumn;
    }
    availableNodesBack += inputs;
    firstInColumn += inputs; //to get its real index in genes vector
    
    srand((unsigned) time(0) + (92 * geneIndex) + Randomizer::getRandomizer()*13); 
    int random = rand();
    int idx = random % (availableNodesBack);
    if(idx < inputs) {
        return this->genes[idx];
    } else {
        idx -= inputs; //go back from first in column
        return this->genes[firstInColumn - idx - 1];
    }
}

/**
 * Outputs must be already assigned! -> use only when mutating, not generating empty circuit
 * @param outputIdx
 */
void MutableCircuit::assignRandomOutput(int outputIdx) {
    int inputs = this->inputsCnt;
    int nodes = this->genes.size() - inputs;
    
    Gene* out;
    string outID;
    
    Gene* newOut;
    string id;
    
    do {
        srand((unsigned) time(0) + 147 * (outputIdx + 7) + Randomizer::getRandomizer()*22); 
        int random = rand();
        int idx = random % nodes;
    
        out = this->outputs[outputIdx];
        outID = out->GetID();
    
        newOut = this->genes[idx + inputs];
        id = newOut->GetID();
    
    } while(id == outID);
    
    int newOutOutputIdx = -1;
    for(int i = 0; i < this->outputs.size(); ++i) {
        if(this->outputs[i] == newOut)
            newOutOutputIdx = i;
    }
    
    out->SetId(id);
    if(newOutOutputIdx > -1)
        this->outputs[newOutOutputIdx] = out;
    
    newOut->SetId(outID);
    this->outputs[outputIdx] = newOut;
    
    this->genotype = "";
}

/**
 * Applies random mutation of 'genes' genes on this MutableCircuit
 */
void MutableCircuit::mutate(int genes) {
    int inputs = this->inputsCnt;
    int genesCnt = this->genes.size() - inputs;
    int outputs = this->outputs.size();
    
    GateMapper* mapper = GateMapper::getInstance();
    for(int i = 0; i < genes; ++i) {
        srand((unsigned) time(0) + (i + 3) * 71 + Randomizer::getRandomizer()*23); 
        int random = rand();
        int idx = random % (genesCnt + outputs);
        if(idx < genesCnt) {
            int geneIdx = inputs + idx;
            Gene* toMutate = this->genes[geneIdx];
            if(toMutate->IsInPhenotype()) {
                this->fitness = -2;
            }
            int in = rand() % (toMutate->GetInputs().size() + 1);
            if(in == toMutate->GetInputs().size()) {                            // change gate
                vector<Gene*> originalInputs = toMutate->GetInputs();
                Gate* gate = mapper->getRandomGate();
                toMutate->SetGate(gate);
                int inCnt = toMutate->GetInputs().size();
                for(int orig_in = 0; orig_in < inCnt; ++orig_in) {
                    if(orig_in < originalInputs.size()) {
                        toMutate->SetInput(orig_in, originalInputs[orig_in]);
                    } else {
                        toMutate->SetInput(orig_in, generateInput(geneIdx));
                    }
                }
            } else { //mutate input
                Gene* input;
                input = generateInput(geneIdx);
                toMutate->SetInput(in, input);
            }
        } else { //mutate output
            idx -= genesCnt;
            assignRandomOutput(idx);
        }
    }
    this->genotype = "";
}
