/* 
 * File:   main.cpp
 * Author: Stanislav Pelák
 */

#define DEBUG_LOG = true

#include "Multilevel/Blif.h"
#include "Generation.h"
#include "CNFmanager.h"
#include "SATsolverAdapter.h"
#include "OutputManager.h"
#include "Randomizer.h"
#include "Configuration.h"
#include "Evolver.h"

#include <ctime> 
using namespace std;

struct Params {
    int r;
    int c;
    int inputs;
    int outputs;
    int l;
    int mutation;
    int mutation_p;
    int lambda;
    int pos;
    int runs;
    char* file;
    char* ofile;
};

/*
 * Help
 */
static void printHelp() {
    cerr << "Usage of application:" << endl;
    cerr << "====================================================================================" << endl;
    cerr << "| Parameter          | Value  | Meaning                       | Required | Default |" << endl;
    cerr << "====================================================================================" << endl;
    cerr << "| -lambda            | number | number of children generated  |     *    |         |" << endl;
    cerr << "|                    |        | in ecah generation            |          |         |" << endl;
    cerr << "------------------------------------------------------------------------------------" << endl;
    cerr << "| -m                 | number | mutation - absolute value     |     *    |         |" << endl;
    cerr << "------------------------------------------------------------------------------------" << endl;
    cerr << "| -eq_children       | 0/1    | 1-only functional individuals |     *    |         |" << endl;
    cerr << "------------------------------------------------------------------------------------" << endl;
    cerr << "| -ran-genotype      | 0/1    | 1-initialize redundant nodes  |     *    |         |" << endl;
    cerr << "|                    |        | randomly                      |          |         |" << endl;
    cerr << "------------------------------------------------------------------------------------" << endl;
    cerr << "| -maxCircuits       | number | maximum generated individuals |     *    |         |" << endl;
    cerr << "------------------------------------------------------------------------------------" << endl;
    cerr << "| -f                 | path   | path to .blif file with refe- |     *    |         |" << endl;
    cerr << "|                    |        | rence circuit                 |          |         |" << endl;
    cerr << "------------------------------------------------------------------------------------" << endl;
    cerr << "| -o                 | path   | path to output folder incl.   |     *    |         |" << endl;
    cerr << "|                    |        | output files' prefixes        |          |         |" << endl;
    cerr << "------------------------------------------------------------------------------------" << endl;
    cerr << "| -r                 | number | number of rows                |          |    1    |" << endl;
    cerr << "------------------------------------------------------------------------------------" << endl;
    cerr << "| -c                 | number | number of columns             |          |    n*   |" << endl;
    cerr << "------------------------------------------------------------------------------------" << endl;
    cerr << "| -l                 | number | levels-back                   |          |    n*   |" << endl;
    cerr << "------------------------------------------------------------------------------------" << endl;
    cerr << "| -pos               | number | starting position of active   |          |    0    |" << endl;
    cerr << "|                    |        | nodes in phenotype            |          |         |" << endl;
    cerr << "------------------------------------------------------------------------------------" << endl;
    cerr << "| -randomIndividuals | true   | generate individuals randomly |          |         |" << endl;
    cerr << "|                    |        | without evolution             |          |         |" << endl;
    cerr << "====================================================================================" << endl << endl;
    cerr << "* - 'n' represents the number of nodes in reference circuit" << endl;
};

static void parseInput(int argc, char** argv, Params* p) {
    p->c = -1;
    p->l = -1;
    p->pos = 0;
    p->r = 1;
    p -> runs = 1;
    
    int required = 0;
    
    if(argc == 0) {
        cerr << "Missing arguments" << endl;
        cerr << endl;
        printHelp();
        exit(1);
    }
        
    try {
        for (int i = 1; i < argc; i++) {
            if ((i + 1) >= argc) {
                cerr << "Invalid arguments: " << argv[i] << endl;
                exit(1);
            }
            string param = argv[i];
            if (param == "-lambda") {
                p->lambda = atoi(argv[++i]);
                required++;
            } else if (param == "-r") {
                p->r = atoi(argv[++i]);
            } else if (param == "-c") {
                p->c = atoi(argv[++i]);
            } else if (param == "-l") {
                p->l = atoi(argv[++i]);
            } else if (param == "-m") {
                p->mutation = atoi(argv[++i]);
                required++;
            } else if (param == "-mp") {
                if(p->mutation > 0)
                    required--;
                required++;
                p->mutation = -1;
                p->mutation_p = atoi(argv[++i]);
            } else if (param == "-f") {
                required++;
                p->file = argv[++i];
            } else if (param == "-o") {
                required++;
                p->ofile = argv[++i];
            } else if (param == "-pos") {
                p->pos = atoi(argv[++i]);
            } else if (param == "-runs") {
                p->runs = atoi(argv[++i]);
            } else if (param == "-smallest") {
                Configuration::SMALLEST_PHENOTYPE = atoi(argv[++i]);
            } else if (param == "-pref") {
                Configuration::NAME_PREFIX = argv[++i];
            } else if (param == "-eq_children") {
                required++;
                Configuration::EQUIV_CHILDREN = (atoi(argv[++i]) == 1);
            } else if (param == "-ran_genotype") {
                required++;
                Configuration::GENERATE_GENOTYPE_RANDOMLY = (atoi(argv[++i]) == 1);
            } else if (param == "-maxCircuits") {
                required++;
                Configuration::MAX_GENERATED_CIRCUITS = atoi(argv[++i]);
            } else if (param == "-randomIndividuals") {
                ++i;
                Configuration::RANDOM_INDIVIDUALS = true;
            } else {
                cout << "Invalid parameter: " << param << endl;
                printHelp();
                exit(1);
            }
        }
    } catch(int e) {
        printHelp();
        exit(1);
    }
    if(required != 7) {
        cout << "Required parameters missing." << endl;
        printHelp();
        exit(1);
    }
};

static string getFilename(string path) {
    int lastIndex = path.find_last_of("/");
    if (!(lastIndex >= 0))
        lastIndex = path.find_last_of("\\");
    string file = path.substr(lastIndex + 1, path.length() - 1);
    int dotIndex = file.find_first_of(".");
    return file.substr(0, dotIndex);
}

int main(int argc, char** argv) {
    //input parameters
    Params params;
    parseInput(argc, argv, &params);
    cout << params.file << endl;

    Blif* b = new Blif();
    b->LoadBlif(params.file);

    //reference circuit
    Circuit* referenceCircuit = new Circuit(b);

    if (params.mutation == -1)
        params.mutation = (int) (b->GetNodes() * params.mutation_p / 100);
    if (params.c == -1)
        params.c = b->GetNodes();
    if (params.l == -1)
        params.l = b->GetNodes();

    cout << "Calculation init:" << endl;
    cout << "r = " << params.r << endl;
    cout << "c = " << params.c << endl;
    cout << "l = " << params.l << endl;
    cout << "pos = " << params.pos << endl;
    cout << "lambda = " << params.lambda << endl;
    cout << "mutation = " << params.mutation << endl;
    cout << "only equivalent individuals = " << Configuration::EQUIV_CHILDREN << endl;
    cout << "file = " << params.file << endl;
    cout << "output file prefix = " << params.ofile << endl;
    cout << "maximum circuits = " << Configuration::MAX_GENERATED_CIRCUITS << endl;

    params.inputs = b->inputs();
    params.outputs = b->outputs();

    string path(params.file);
    string ofile(params.ofile);
    OutputManager::init(ofile + getFilename(path));

    Evolver::evolveBasic(path, ofile, referenceCircuit, params.r, params.c, params.l, params.lambda, params.mutation, params.inputs, params.outputs, params.pos);
}














